/*
	This file includes weak (i.e. optional) functions to perform SoC related
	initialization. They are:
	1). _nds32_init_mem():
		Executed before C language initialization to make memory
		ready so that program data can be initialized. An example
		is to initialize DRAM.
		Since this is called before C initialization, please
		use provided macros to avoid problems.
	2). __soc_init():
		Further SoC intialization. Called after C language
		initialization, so it is a typical C function.
*/

#include "nds32_intrinsic.h"
#include "config.h"
#ifdef CFG_16MB
#include "ag101_16mb.h"
#else
#ifdef CFG_AG102
#include "ag102.h"
#else
#include "ag101.h"
#endif
#endif

#ifdef CFG_LLINIT
/* This must be a leave function, no child funcion. */
void _nds32_init_mem(void) __attribute__((no_prologue, optimize("Os")));
void _nds32_init_mem(void) //The function is weak (optional)
{
#ifdef CFG_AG102	// AG102
#ifndef CFG_NORAM
	/* initialize DDR2 controller */
	do {
		NDS32_REG_WRITE(DDR2C_BASE + DDR2C_DCR_OFFSET, 0x000025cc);     // SDRAM configure
		NDS32_REG_WRITE(DDR2C_BASE + DDR2C_DTAR_OFFSET, 0x00100000);    // Data training addr.
		NDS32_REG_WRITE(DDR2C_BASE + DDR2C_DRR_OFFSET, 0x00034812);     // Set refresh
		NDS32_REG_WRITE(DDR2C_BASE + DDR2C_MR_OFFSET, 0x00000852);      // Set mode
		NDS32_REG_WRITE(DDR2C_BASE + DDR2C_IOCR_OFFSET, 0x0000000f);    // Set IOCR

		/*
		 * Enable host ports and trigger initialization
		 */
		NDS32_REG_WRITE(DDR2C_BASE + DDR2C_CCR_OFFSET, 0x80020000);

		/* Wait for ddr init state to be set */
		__nds32__isb();

		/* Wait until the config initialization is finish */
		while(NDS32_REG_READ(DDR2C_BASE + DDR2C_CSR_OFFSET) >> 23);

		/*
		 * Enable host ports and trigger the data training
		 */
		NDS32_REG_WRITE(DDR2C_BASE + DDR2C_CCR_OFFSET, 0x40020004);

		/* Wait for ddr init state to be set */
		__nds32__isb();

		/* Wait until the ddr data trainning is complete */
		while(NDS32_REG_READ(DDR2C_BASE + DDR2C_CSR_OFFSET) >> 23);

		register unsigned int v;
		v = NDS32_REG_READ(DDR2C_BASE + DDR2C_CSR_OFFSET);
		NDS32_REG_WRITE(DDR2C_BASE + DDR2C_CSR_OFFSET, 0x00ffffff);
		if(!(v >> 20))
			break;
	} while(1);
#endif	// CFG_NORAM
#ifdef CFG_REMAP
	extern char _edata;

	/* relocation, copy ROM code to SDRAM (0x10000000) */
	NDS32_FMEMCPY(ORIG_RAM_BASE, 0x0, &_edata);

	/* set remap bit */
	NDS32_SET_BITS(AHBC_BASE + INTC_OFFSET, 0x1); //remap!
	__nds32__isb();
#else	// XIP
	/* The SDRAM is mapped to 0x10000000 for use. */
	NDS32_REG_WRITE(AHBC_BASE + DEVICE4_OFFSET, 0x00000000 | 0x80000); //ROM (base,size)=(0M,256M)
	NDS32_REG_WRITE(AHBC_BASE + DEVICE6_OFFSET, 0x10000000 | 0x80000); //RAM (base,size)=(256M,256M)

#if defined(CFG_NORAM) && !defined(CFG_SIMU)
	/* System without SDRAM. Use data local memory as system memory. */
	extern char __data_start;
	register unsigned int dlmsize;

	dlmsize = 0x1000 << ((__nds32__mfsr(NDS32_SR_DLMB) >> 1) & 0xf);

	/* Set DLM base to .data start address and enable it */
	__nds32__mtsr((unsigned)&__data_start|1, NDS32_SR_DLMB);
	__nds32__dsb();

	/* Update stack pointer to end of DLM
	 * We suppose the .data + .bss + stack less then DLM size */
	__nds32__set_current_sp((unsigned)&__data_start + dlmsize);
#endif
#endif	// CFG_REMAP

#else	// AG101 or AG101P
	//initialize memory controllers of AG101 before first use
	/* static memory bank 0 timing parameter register */
	NDS32_REG_WRITE(SMC_BASE + BANK0TPR_OFFSET, SMC_BANK0TPR_INIT);

	/* PMU - PLL/DLL Control Register 0 - Enable DLL */
	NDS32_SET_BITS(PMUC_BASE + PLLDLLCR0_OFFSET, 0x00010000);
#ifdef CFG_AG101
	NDS32_SET_BITS(PMUC_BASE + PLLDLLCR0_OFFSET, 0x00e00000);
	NDS32_SET_BITS(PMUC_BASE + MFPSR_OFFSET, 0x00000010);
	NDS32_SET_BITS(PMUC_BASE + SHTCR_OFFSET, 0x0007f000);
#endif

	/* SDRAM controller - set timing parameter */
	NDS32_REG_WRITE(SDRAMC_BASE + PARM1_OFFSET, SDRAMC_PARM1_INIT);
	NDS32_REG_WRITE(SDRAMC_BASE + PARM2_OFFSET, SDRAMC_PARM2_INIT);
	/* SDRAM controller - set config register */
	NDS32_REG_WRITE(SDRAMC_BASE + CONFIG1_OFFSET, SDRAMC_CONFIG1_INIT);
	/* refer to ATFSDMC appendix, and boot.S of bootcode */
	NDS32_REG_WRITE(SDRAMC_BASE + CONFIG2_OFFSET, SDRAMC_CONFIG2_INIT); /* 0x10 */
	NDS32_REG_WAIT4_BIT_OFF(SDRAMC_BASE + CONFIG2_OFFSET, 0x1c);
	NDS32_REG_WRITE(SDRAMC_BASE + CONFIG2_OFFSET, 0x4);
	NDS32_REG_WAIT4_BIT_OFF(SDRAMC_BASE + CONFIG2_OFFSET, 0x1c);
	NDS32_REG_WRITE(SDRAMC_BASE + CONFIG2_OFFSET, 0x8);
	NDS32_REG_WAIT4_BIT_OFF(SDRAMC_BASE + CONFIG2_OFFSET, 0x1c);

#ifdef CFG_REMAP
	/* remap */
	extern char _edata;
	/* relocation, copy ROM code to SDRAM (0x10000000/4GB or 0x800000/16MB) */
#ifdef CFG_16MB
	NDS32_REG_WRITE(SDRAMC_BASE + 0x10, 0x00001008); //8MB
#else
	NDS32_REG_WRITE(SDRAMC_BASE + 0x10, 0x00001100); //256MB
#endif

	NDS32_FMEMCPY(ORIG_RAM_BASE, 0x0, &_edata);

	/* Remapping */
	/* Bank Enable, BASE = 0x100 for haddr[31:20] of the AHB address bus. */
	/* External Bank Configuration Registers */
	NDS32_REG_WRITE(SDRAMC_BASE + 0x14, 0x0);
	NDS32_REG_WRITE(SDRAMC_BASE + 0x18, 0x0);
	NDS32_REG_WRITE(SDRAMC_BASE + 0x1c, 0x0);
	/* Bank Enable, BASE = 0x0 for haddr[31:20] of the AHB address bus. */
	NDS32_REG_WRITE(SDRAMC_BASE + 0x10, 0x1000);
	NDS32_SET_BITS(AHBC_BASE + INTC_OFFSET, 0x1); //remap!
	__nds32__isb();

#ifdef CFG_16MB
	NDS32_REG_WRITE(AHBC_BASE + 0x10, 0x0080000b); //ROM (base,size)=(8M,1M)
	NDS32_REG_WRITE(AHBC_BASE + 0x18, 0x0000000e); //RAM (base,size)=(0M,8M)
	NDS32_REG_WRITE(SMC_BASE, 0x10800052); //ROM: 8M~
#endif

	/* When writing a 1 to this bit, the base/size configuration
	 * of AHB slaves 4 (ROM) and 6 (SDRAM) will be interchanged. */
#ifdef CFG_AG101
	NDS32_REG_WRITE(SMC_BASE + 0x8, 0x10400062);
	NDS32_REG_WRITE(SMC_BASE + 0xc, 0x000ff3ff);
#else
#ifndef CFG_16MB
	/*64 MB, Memory data width is 32 */
	NDS32_REG_WRITE(SMC_BASE, 0x10000062); //ROM: 2G~
#endif
#endif
#else //XIP, !defined(CFG_REMAP)
#ifdef CFG_16MB
	NDS32_REG_WRITE(AHBC_BASE + 0x10, 0x0000000b); //ROM (base,size)=(0M,1M)
	NDS32_REG_WRITE(AHBC_BASE + 0x18, 0x0080000d); //RAM (base,size)=(8M,4M)
	NDS32_REG_WRITE(SMC_BASE, 0x10000052); //ROM: 0M~
#endif
#if defined(CFG_NORAM) && !defined(CFG_SIMU)
	/* System without SDRAM. Use data local memory as system memory. */
	extern char __data_start;
	register unsigned int dlmsize;

	dlmsize = 0x1000 << ((__nds32__mfsr(NDS32_SR_DLMB) >> 1) & 0xf);

	/* Set DLM base to .data start address and enable it */
	__nds32__mtsr((unsigned)&__data_start|1, NDS32_SR_DLMB);
	__nds32__dsb();

	/* Update stack pointer to end of DLM
	 * We suppose the .data + .bss + stack less then DLM size */
	__nds32__set_current_sp((unsigned)&__data_start + dlmsize);
#endif
#endif //#ifdef CFG_REMAP
#endif //#ifdef CFG_AG102
}
#endif //#ifdef CFG_LLINIT

void __soc_init() //The function is weak (optional)
{
#ifdef CFG_LLINIT
	void uart_init(void);
	uart_init();
#endif
}
